/*
firestring.c - FireString replacement string functions
Copyright (C) 2000 Ian Gulliver

This program is free software; you can redistribute it and/or modify
it under the terms of version 2 of the GNU General Public License as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Parts of this library are taken from dstr and dconf, part of LibD:
* Copyright (C) 2001-2002  Andres Salomon <dilinger@mp3revolution.net>

*/

#include "firestring.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>

void *firestring_malloc(const size_t size) {
	char *output;
	output = malloc(size);
	if (output == NULL) {
		perror("malloc");
		exit(EXIT_FAILURE);
	}
	return output;
}

void *firestring_realloc(void *old, const size_t new) {
	char *output;
	output = realloc(old,new);
	if (output == NULL) {
		perror("realloc");
		exit(EXIT_FAILURE);
	}
	return output;
}

char *firestring_strdup(const char * const input) {
	char *output;
	size_t s;
	if (input == NULL)
		return NULL;
	s = strlen(input) + 1;
	output = firestring_malloc(s);
	memcpy(output,input,s);
	return(output);
}

void firestring_strncpy(char * const to, const char * const from, const size_t size) {
	strncpy(to,from,size);
	to[size - 1]= '\0';
	return;
}

void firestring_strncat(char * const to, const char * const from, const size_t size) {
	size_t l;
	l = strlen(to);
	firestring_strncpy(&to[l],from,size - l);
	return;
}

static char *shownum_unsigned(unsigned long long m) {
	static char numbuf[128];
	int power; /* smallest power of 10 greater than m */
	long long multiple = 1;
	int r;
	int i = 0;

	for (power = 0; multiple <= m; power++)
		multiple *= 10;

	for (;power > 0; power--) {
		multiple /= 10;
		r = m / multiple;
		numbuf[i++] = '0' + r;
		m -= r * multiple;
	}
	numbuf[i] = '\0';
	return numbuf;
}

static char *shownum_signed(long long n) {
	static char numbuf[128];
	unsigned long long m;
	int i = 0;

#define ABS(n) ((n < 0) ? n * -1 : n);
	m = (unsigned long long) ABS(n);
	if ((long long) m != n)
		numbuf[i++] = '-';
	firestring_strncpy(&numbuf[i],shownum_unsigned(m),127);
	return numbuf;
}

int firestring_snprintf(char * out, const size_t size, const char * const format, ...) {
	va_list ap;
	char *numresult;
	size_t f,o = 0,fl,tl,ml;
	char *tempchr;
	int b = 0;

	fl = strlen(format);
	ml = size - 1;

#define NUMBER tl = strlen(numresult); \
	if (tl + o >= ml) \
		b = 1; \
	else { \
		memcpy(&out[o],numresult,tl); \
		o += tl; \
	}

#define SIGNED_NUMBER(type) \
	numresult = shownum_signed((long long) va_arg(ap,type)); \
	NUMBER

#define UNSIGNED_NUMBER(type) \
	numresult = shownum_unsigned((unsigned long long) va_arg(ap,type)); \
	NUMBER

	va_start(ap,format);
	for (f = 0; f < fl && o < ml && b == 0; f++) {
		if (format[f] == '%') {
			switch(format[++f]) {
				case 's':
					tempchr = va_arg(ap,char *);
					if (tempchr == NULL)
						tempchr = "(null)";
					tl = strlen(tempchr);
					if (tl + o >= ml)
						b = 1;
					else {
						memcpy(&out[o],tempchr,tl);
						o += tl;
					}
					break;
				case 'd': /* signed int */
					SIGNED_NUMBER(int)
					break;
				case 'l': /* signed long */
					SIGNED_NUMBER(long)
					break;
				case 'u': /* unsigned int */
					UNSIGNED_NUMBER(unsigned int)
					break;
				case 'y': /* unsigned long */
					UNSIGNED_NUMBER(unsigned long)
					break;
				case 'g': /* signed long long */
					SIGNED_NUMBER(long long)
					break;
				case 'o': /* unsigned long long */
					UNSIGNED_NUMBER(unsigned long long)
					break;
				case '%':
					out[o++] = '%';
					break;
			}
		} else
			out[o++] = format[f];
	}
	out[o] = '\0';
	return o;
}

int firestring_strncasecmp(const char * const s1, const char * const s2, const size_t n) {
	size_t s;
	for (s = 0; s < n; s++) {
		if (tolower((unsigned char) s1[s]) != tolower((unsigned char) s2[s])) {
			if (tolower((unsigned char) s1[s]) < tolower((unsigned char) s2[s]))
				return -1;
			else
				return 1;
		}
		if (s1[s] == '\0')
			return 0;
	}
	return 0;
}

int firestring_strcasecmp(const char * const s1, const char * const s2) {
	size_t s;
	char c1, c2;
	for (s = 0; (s1[s] != '\0') && (s2[s] != '\0') && (tolower((unsigned char) s1[s]) == tolower((unsigned char) s2[s])); s++);

	c1 = tolower((unsigned char) s1[s]);
	c2 = tolower((unsigned char) s2[s]);
	if (c1 == c2)
		return 0;
	else if (c1 < c2)
		return -1;
	else
		return 1;
}

char *firestring_concat (const char *s, ...) {
	const char *curr;
	size_t len = 0;
	va_list va;
	char *ret = NULL;

	/* get length */
	va_start(va,s);
	curr = s;
	while (curr != NULL) {
		len += strlen(curr);
		curr = va_arg(va, const char *);
	}
	va_end(va);

	/* allocate string */
	if (len) {
		ret = firestring_malloc(len + 1);
		*ret = '\0';
	}

	/* copy */
	va_start(va,s);
	curr = s;
	while (curr != NULL) {
		strcat(ret,curr);
		curr = va_arg(va, const char *);
	}
	va_end(va);

	return ret;
}

char *firestring_chomp(char *s) {
	char *ret = s;
	if (s != NULL) {
		s = &s[strlen(s) - 1];
		while (isspace(*s)) {
			*s = '\0';
			if (s == ret)
				break;
			s--;
		}
	}
	return ret;
}

char *firestring_chug(char *s) {
	if (s != NULL) {
		while (isspace(*s))
			s++;
	}
	return s;
}

static char *try_escaped_newline(char *s) {
	char *esc, *end;
	
	/* check for terminating '\\' */
	esc = end = strrchr (s, '\\');
	if (end) {
		end++;

		/* check if the only thing separating the '\\' and the end
		 * of the string is just space */
		while (*end && isspace (*end))
			end++;
		if (*end == '\0')
			*esc = '\0';
	}

	return s;
}

struct firestring_conf_t *firestring_conf_parse(const char *filename) {
	struct firestring_conf_t *list = NULL;
	FILE *fp;
	char buf[512];
	char *var, *val, *end;
	char *prev_var = NULL, *prev_val = NULL;

	fp = fopen(filename, "r");
	if (!fp)
		return NULL;

	while (fgets(buf, sizeof (buf), fp) != NULL) {
		/* break into variable and value */
		if (prev_var == NULL) {
			var = firestring_chug(buf);
			if (*var == '#')
				continue;	/* comment */

			val = strchr(var,'=');
			if (val == NULL)
				continue;	/* malformed line */
			*val++ = '\0';

			val = firestring_chug(val);
			var = firestring_chomp(var);
		} else {
			/* continuing multi-line */
			var = prev_var;
			val = firestring_concat(prev_val, buf, NULL);
			free(prev_val);
			prev_val = NULL;
		}

		/* handle quoted and multi-line values */
		if (*val == '"') {
			end = strrchr(val, '"');
			if (end == val || *(end - 1) == '\\') {
				val = try_escaped_newline(val);
				prev_var = firestring_strdup(var);
				prev_val = firestring_strdup(val);
				continue;
			}
			val++;
			*end = '\0';
		}
		else
			val = firestring_chomp(val);

		/* save variable/value to list */
		list = firestring_conf_add(list, var, val);
		if (prev_var != NULL) {
			free(prev_var);
			prev_var = NULL;
		}
	}

	fclose (fp);
	return list;
}

struct firestring_conf_t *firestring_conf_add(struct firestring_conf_t *next, const char *var, const char *value) {
	struct firestring_conf_t *conf;

	if (var != NULL && value != NULL) {
		conf = firestring_malloc(sizeof(*conf));
		conf->next = next;
		conf->var = firestring_strdup(var);
		conf->value = firestring_strdup(value);
	} else
		conf = next;

	return conf;
}

const char *firestring_conf_find(struct firestring_conf_t *config, const char *var) {
	while (config != NULL) {
		if (firestring_strcasecmp(config->var, var) == 0)
			return config->value;
		config = config->next;
	}

	return NULL;
}

const char *firestring_conf_find_next(struct firestring_conf_t *config, const char *var, const char *prev) {
	int canreturn = 0;

	if (prev == NULL)
		canreturn = 1;

	while (config != NULL) {
		if (firestring_strcasecmp(config->var, var) == 0) {
			if (canreturn == 1)
				return config->value;
			else if (config->value == prev)
				canreturn = 1;
		}
		config = config->next;
	}

	return NULL;
}

void firestring_conf_free(struct firestring_conf_t *config) {
	struct firestring_conf_t *next;

	while (config != NULL) {
		next = config->next;
		free(config->var);
		free(config->value);
		free(config);
		config = next;
	}
}
